from typing import Generic, TypeVar, cast, Type, Optional, Any, TYPE_CHECKING

from lacia.standard import JsonAst
from lacia.types import RpcMessage, JsonRpcCode
from lacia.standard.abcbase import BaseDataTrans
from lacia.exception import JsonRpcRuntimeException

if TYPE_CHECKING:
    from lacia.core.core import JsonRpc

T = TypeVar("T", bound=BaseDataTrans)

Schema = TypeVar("Schema")

def with_schema(obj, schema: Type[Schema]) -> Schema:
    return cast(schema, obj)

class BaseProxy(Generic[T]):
    _jsonrpc: str
    _obj: T

    def __await__(self):
        ...

class ProxyObj(BaseProxy):
    _jsonrpc: str = "jsonast"

    def __init__(self, core: Optional["JsonRpc[JsonAst]"] = None, name: Optional[str] = None):
        self._aiter: Optional["ResultProxy"] = None
        self._obj = None # type: ignore
        self._core = core
        self._name = name

    def __getattr__(self, name: str) -> "ProxyObj":
        obj = JsonAst( 
            obj=self._obj,
            method="__getattr__",
            args=(name,),
            kwargs={},
        )
        self._obj = obj
        return self

    def __call__(self, *args, **kwargs) -> "ProxyObj":

        n_args = []
        n_kwargs = {}
        for i in args:
            if isinstance(i, ProxyObj):
                n_args.append(i._obj)
            else:
                n_args.append(i)
        
        for k,v in kwargs.items():
            if isinstance(v, ProxyObj):
                n_kwargs[k] = v
            else:
                n_kwargs[k] = v

        obj = JsonAst(
            obj=self._obj,
            method="__call__",
            args=tuple(n_args),
            kwargs=n_kwargs,
        )
        self._obj = obj
        return self

    def __next__(self):
        ...

    def __aiter__(self):
        obj = JsonAst(
            obj=self._obj,
            method="__aiter__",
            args=(),
            kwargs={}
        )
        self._obj = obj

        return self
    
    def __anext_proxy__(self):
        obj = JsonAst(
            obj=self._obj,
            method="__anext__",
            args=(),
            kwargs={}
        )
        self._obj = obj

        return self

    async def __anext__(self):
        if self._core is None:
            raise TypeError("ProxyObj is not bind to JsonRpc")
        if not self._aiter:
            data = await self._core.run(self)
            self._aiter = data
            r = await data.__anext__()
            return r
        else:
            r = await self._aiter.__anext__()
            return r

    def __await__(self):
        if self._core is None:
            raise TypeError("ProxyObj is not bind to JsonRpc")
        if not self._core._client is None:
            data = yield from self._core.run(self).__await__()
        elif not self._core._server is None:
            if self._name is None:
                raise JsonRpcRuntimeException("client name is None")
            data = yield from self._core.reverse_run(self._name, self).__await__()
        else:
            raise JsonRpcRuntimeException("server and client are None")
        self._obj = None
        return data.visions

class ResultProxy(BaseProxy):

    def __init__(self, result: RpcMessage, core: Optional["JsonRpc[JsonAst]"] = None):
        self._core = core
        self._result = result
        self._obj = None

    @property
    def visions(self):
        if self._result.error:
            if self._result.error_code == JsonRpcCode.StopAsyncIterationError:
                raise StopAsyncIteration
            raise JsonRpcRuntimeException(self._result.error)
        return self._result.result

    def __getattr__(self, name: str) -> "ProxyObj":
        return getattr(getattr(ProxyObj(self._core), self._result.id), name) # type: ignore

    async def __aiter__(self):
        return self

    async def __anext__(self):
        if self._core is None:
            raise TypeError("ProxyObj is not bind to JsonRpc")
        obj = getattr(ProxyObj(self._core), self._result.id).__anext_proxy__()

        data = await self._core.run(obj)
        return data.visions




from abc import ABC, abstractmethod, abstractclassmethod

from typing import TypeVar, Generic, Type, Any, Dict, NamedTuple, Hashable
from typing_extensions import Self

T = TypeVar("T")
S = TypeVar("S")

class Namespace(NamedTuple):

    builtins: Dict[str, Any] = {}
    globals: Dict[str, Any] = {}
    locals: Dict[Hashable, Dict[str, Any]] = {}

    def __getitem__(self, key: T) -> Dict[str, Any]:

        return {
            **self.builtins,
            **self.globals,
            **self.locals.get(key, {})
        }      

class BaseDataTrans(ABC, Generic[T]):

    @abstractclassmethod # type: ignore
    def loads(cls, obj: T) -> Self:
        ...
    
    @abstractmethod
    def dumps(self) -> T:
        ...

class BaseRunTime(ABC, Generic[S]):

    def __init__(self, namespace: Dict[str, Any]):
        self.namespace = namespace

    @abstractmethod
    async def run(self, obj: S) -> Any:
        ...

class BaseStandard(ABC, Generic[T, S]):
    
    datatrans: Type[BaseDataTrans[T]]
    runtime: Type[BaseRunTime[S]]



import asyncio
import json
from uuid import uuid4
from contextvars import ContextVar
from typing import Dict, Any, Optional, TypeVar, Generic

from nest_asyncio import apply as nest_apply

from lacia.core.abcbase import BaseJsonRpc
from lacia.core.proxy import BaseProxy, ResultProxy, ProxyObj
from lacia.network.abcbase import BaseServer, BaseClient
from lacia.standard.abcbase import BaseDataTrans, Namespace
from lacia.standard.execute import Standard
from lacia.logger import logger
from lacia.types import RpcMessage
from lacia.exception import JsonRpcInitException

T = TypeVar("T")

class Context:
    websocket: ContextVar = ContextVar("websocket")
    name: ContextVar[str] = ContextVar("name")
    namespace: ContextVar[Namespace] = ContextVar("namespace")
    rpc: ContextVar = ContextVar("rpc")

class JsonRpc(BaseJsonRpc, Generic[T]):
    
    def __init__(
        self,
        name: str,
        execer: bool = True,
        namespace: Optional[Dict[str, Any]] = None,
        token: Optional[str] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        debug: bool = False,
 
    ) -> None:
        self._name = name
        self._execer = execer
        self._namespace = Namespace(
            globals=namespace if namespace else {},
        )
        self._token = token
        self._loop = loop
        self._debug = debug
        self._uuid = str(uuid4())

        self._wait_remote: Dict[str, asyncio.Event] = {}
        self._wait_result: Dict[str, ResultProxy] = {}

        self._standard = Standard()

    def add_namespace(self, namespace: Dict[str, Any]) -> None:
        self._namespace.globals.update(namespace)

    async def run_client(self, client: BaseClient) -> None:
        self._standard.init_standard()
        self._client = client
        self._loop = self._loop or asyncio.get_event_loop()
        await client.start()
        if self._loop:
            self._loop.create_task(self._listening_server(self._client.ws))
            logger.info("run client")
    
    async def run_server(self, server: BaseServer) -> None:
        self._loop = self._loop or asyncio.get_event_loop()
        nest_apply(self._loop)
        self._standard.init_standard()
        self._server = server
        self._server.on("connect", self._listening_client)
        self._server.on("disconnect", self.on_server_close)
        logger.info("run server")
        await server.start()
    
    async def _listening_client(self, websocket: T):
        logger.info("listening client")
        Context.websocket.set(websocket)
        Context.namespace.set(self._namespace)
        Context.rpc.set(self)

        event = asyncio.Event()

        def rpc_auto_register(name, token):
            if self._server is not None:
                self._server.active_connections.set_name_ws(name, websocket)
            Context.name.set(name)
            if token == self._token:
                event.set()
                return
            raise JsonRpcInitException("rpc_auto_register fail")

        if self._namespace.locals.get(websocket) is None:
            self._namespace.locals[websocket] = {}
        if self._server:
            self._namespace.locals[websocket]["rpc_auto_register"] = rpc_auto_register

        qmgs = asyncio.Queue()

        if self._loop:
            self._loop.create_task(self._client_auth(event, qmgs, websocket))

        if self._server is not None:
            async for message in self._server.iter_json(websocket):
                logger.info(f"receive: {message}")
                msg = RpcMessage(message)
                if msg.is_request and self._execer and self._loop:
                    if msg.is_auth:
                        self._loop.create_task(self._s_execute(websocket, msg))
                    else:
                        qmgs.put_nowait(msg)
                elif msg.is_response:
                    rmsg = ResultProxy(msg, core=self) # type: ignore
                    self._wait_result[msg.id] = rmsg
                    self._wait_remote[msg.id].set()
                    
        else:
            raise JsonRpcInitException("server is None")

    async def _listening_server(self, websocket: T):
        logger.info("listening server")

        if self._client is not None:
            
            if self._loop:
                self._loop.create_task(self.run(ProxyObj().rpc_auto_register(self._name, self._token)))

            async for message in self._client.iter_json():
                logger.info(f"receive: {message}")
                msg = RpcMessage(message)

                if msg.is_request and self._execer and self._loop:
                    self._loop.create_task(self._c_execute(websocket, msg))
                elif msg.is_response:
                    rmsg = ResultProxy(msg, core=self) # type: ignore
                    self._wait_result[msg.id] = rmsg
                    self._wait_remote[msg.id].set()
            return
        else:
            raise JsonRpcInitException("client is None")
    
    async def _s_execute(self, websocket: T, message: RpcMessage): 

        result, error = await self._standard.rpc_request(message.data, self._namespace[websocket])

        if error is None:
            if websocket not in self._namespace.locals:
                self._namespace.locals[websocket] = {}
            self._namespace.locals[websocket][message.id] = result # TODO 内存泄漏
            msg = {
                "jsonrpc": message.jsonrpc,
                "id": message.id,
                "result": self._pretreatment(result)
            }
        else:
            msg = {
                "jsonrpc": message.jsonrpc,
                "id": message.id,
                "error": error
            }
        if self._server is not None:
            logger.info(f"send: {msg}")
            
            await self._server.send_json(websocket, msg)
    
    async def _c_execute(self, websocket: T, message: RpcMessage):

        result, error = await self._standard.rpc_request(message.data, self._namespace[websocket])

        if error is None:
            if websocket not in self._namespace.locals:
                self._namespace.locals[websocket] = {}
            self._namespace.locals[websocket][message.id] = result # TODO 内存泄漏
            msg = {
                "jsonrpc": message.jsonrpc,
                "id": message.id,
                "result": self._pretreatment(result)
            }
        else:
            msg = {
                "jsonrpc": message.jsonrpc,
                "id": message.id,
                "error": error
            }

        if self._client is not None:
            logger.info(f"send: {msg}")
            logger.info(self._namespace)
            await self._client.send_json(msg) 

    def on_client_close(self, websocket: T):
        del self._namespace.locals[websocket]
    
    def on_server_close(self, websocket: T):
        del self._namespace.locals[websocket]

    async def run(self, proxy: BaseProxy[BaseDataTrans]):
        uuid_str = str(uuid4())
        if proxy._obj is None:
            raise JsonRpcInitException("proxy._obj is None")
        data = proxy._obj.dumps()
        event = asyncio.Event()

        self._wait_remote[uuid_str] = event

        msg = {
            "jsonrpc": proxy._jsonrpc,
            "id": uuid_str,
            "method": data,
        }

        if self._client is not None:
            logger.info(f"send: {msg}")
            await self._client.send_json(msg)
        else:
            raise JsonRpcInitException("server and client are None")

        await event.wait()

        return self._wait_result.pop(uuid_str)
 
    async def reverse_run(self, name: str, proxy: BaseProxy[BaseDataTrans]):
        uuid_str = str(uuid4())
        if proxy._obj is None:
            raise JsonRpcInitException("proxy._obj is None")
        data = proxy._obj.dumps()
        event = asyncio.Event()

        self._wait_remote[uuid_str] = event

        msg = {
            "jsonrpc": proxy._jsonrpc,
            "id": uuid_str,
            "method": data,
        }

        if self._server is not None:
            logger.info(f"send: {msg}")
            await self._server.send_json(self._server.active_connections.get_ws(name), msg)
        else:
            raise JsonRpcInitException("server and client are None")

        await event.wait()

        return self._wait_result.pop(uuid_str)

    async def _client_auth(self, event: asyncio.Event, qmgs: asyncio.Queue, websocket: T):
        await event.wait()
        if self._server is not None:
            Context.name.set(self._server.active_connections.get_name(websocket))
        while True:
            msg = await qmgs.get()
            await self._s_execute(websocket, msg)

    def _pretreatment(self, data: Any) -> Any:
        try:
            json.dumps(data)
            return data
        except TypeError:
            return str(data)

    def is_server(self) -> bool:
        if self._server is None and self._client is None:
            raise JsonRpcInitException("server and client are None")
        elif self._client is None and self._server is not None:
            return True
        return False

    @property
    def jsonast(self):
        return ProxyObj(self) # type: ignore




import asyncio
from typing import Dict, Any, Type, TYPE_CHECKING

from lacia.standard.abcbase import BaseRunTime, BaseStandard, Namespace
from lacia.standard.jsonast.impl import JsonAst

if TYPE_CHECKING:
    from lacia.core.proxy import ProxyObj

class RunTime(BaseRunTime[JsonAst]):
    
    def __init__(self, namespace: Dict[str, Any]):
        self.namespace = namespace

    async def run(self, ast: JsonAst):
        if isinstance(ast, JsonAst):
            if isinstance(ast.obj, JsonAst):
                obj = await self.run(ast.obj)
            elif ast.obj is None:
                if not ast.args is None:
                    if len(ast.args) == 1 and isinstance(ast.args[0], str):
                        obj = self.namespace[ast.args[0]]
                    else:
                        raise TypeError(f"obj type error: {ast.obj}")
                else:
                    raise TypeError(f"obj type error: {ast.obj}")    
            elif isinstance(ast.obj, str):
                obj = self.namespace[ast.obj]
            else:
                raise TypeError(f"obj type error: {ast.obj}")
            if ast.method is None or ast.obj is None:
                return obj
            if not ast.args is None:
                if len(ast.args) == 1 and ast.method == "__getattr__" and isinstance(ast.args[0], str):
                    return getattr(obj, ast.args[0])
            func = getattr(obj, ast.method)
            if ast.args is None and ast.kwargs is None:
                return func
            if ast.method == "__anext__":
                return await func()
            
            args = []
            kwargs = {}
            for i in ast.args or ():
                args.append(await self.run(i))
            if ast.kwargs:
                for k,v in ast.kwargs.items():
                    kwargs[k] = await self.run(v)
            args = tuple(args)
            if ast.method == "__call__" and asyncio.iscoroutinefunction(obj):
                return await func(*args, **kwargs)
            if asyncio.iscoroutinefunction(func):
                return await func(*args, **kwargs)
            return func(*args, **kwargs)
        else:
            return ast

class Standard(BaseStandard[dict, JsonAst]):

    _jsonrpc = "jsonast"

    datatrans: Type[JsonAst] = JsonAst
    runtime: Type[RunTime] = RunTime




from __future__ import annotations

from enum import Enum

from typing import Dict, Any

Message = Dict[str, Any]

class JsonRpcCode(int, Enum):
    ParseError = -32700             
    InvalidRequest = -32600
    MethodNotFound = -32601
    InvalidParams = -32602
    InternalError = -32603
    ServerError = -32000
    StopAsyncIterationError = -32099

class RpcMessage:

    def __init__(self, data: Dict[str, Any]):
        self.data = data
    
    @property
    def jsonrpc(self) -> str:
        return self.data.get("jsonrpc", None)
    
    @property
    def id(self) -> Any:
        return self.data.get("id", None)

    @property
    def method(self):
        return self.data.get("method", {})

    @property
    def result(self) -> Any:
        return self.data.get("result", None)
    
    @property
    def error(self) -> Any:
        return self.data.get("error", None)
    
    @property
    def error_code(self) -> int:
        return self.data["error"]["code"]

    @property
    def error_msg(self) -> str:
        return self.data["error"]["message"]
    
    @property
    def is_error(self) -> bool:
        return "error" in self.data

    @property
    def is_request(self) -> bool:
        return "method" in self.data
    
    @property
    def is_response(self) -> bool:
        return "result" in self.data or "error" in self.data

    @property
    def is_auth(self) -> bool:

        return all(
            [
                self.method["obj"] == {'obj': None, 'method': '__getattr__', 'args': ['rpc_auto_register'], 'kwargs': {}},
            ]
        )
